/*
 *
 *
 *  (C) 1995 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

/*#include "mpid.h" */
/*#include "mpiddev.h" */
/*#include "reqalloc.h" */
#include "hot_stuff.h"
#include "queue.h"


/* 
  This file contains the routines to handle receiving a message
 */

/***************************************************************************/
/*
 * Despite the apparent symmetry, receives are fundamentally different from
 * sends.  All receives happen by processing an incoming item of information
 * and checking it against known receives.
 *
 * Eventually, we may want to make RecvContig a special case (as in the
 * first generation ADI) to avoid the routine calls.
 */
/***************************************************************************/
/* Does this need to return msgrep if heterogeneous? */

/* 
 * Error handling needs additional work.  Is a request that has detected an
 * error complete?  Who sets it?  What if the error is ERR_TRUNCATE (where
 * we'd like to make everything work to the limit of the buffer)?
 */

void MPID_RecvContig( 
	struct MPIR_COMMUNICATOR *comm_ptr, 
	void *buf, 
	int maxlen, 
	int src_lrank, 
	int tag, 
	int context_id, 
	MPI_Status *status, 
	int *error_code )
{
    MPIR_RHANDLE *rhandle;
    MPI_Request  request;

    MPID_RecvAlloc( rhandle );
    request = (MPI_Request)(rhandle);
    MPID_Request_init( rhandle , MPIR_RECV );
    DEBUG_INIT_STRUCT(request,sizeof(*rhandle));
    
    *error_code = 0;
    MPID_IrecvContig( comm_ptr, buf, maxlen, src_lrank, tag, context_id, 
		      request, error_code );
    if (!*error_code) {
      MPID_RecvComplete( request, status, error_code); 
    }
    
    MPID_RecvFree( rhandle ); 
}


void MPID_IrecvContig( 
	struct MPIR_COMMUNICATOR *comm_ptr, 
	void *buf, 
	int maxlen, 
	int src_lrank, 
	int tag, 
	int context_id, 
	MPI_Request request, 
	int *error_code )
{
    MPIR_RHANDLE *dmpi_unexpected;
    MPIR_RHANDLE *rhandle = &(request->rhandle);
    
    DEBUG_PRINT_ARGS("R starting IrecvContig");
    
    /* The one error test that makes sense here */
    if (buf == NULL && maxlen > 0) {
      *error_code = MPI_ERR_BUFFER;
      return;
    }
    
    /* 
       At this time, we check to see if the message has already been received.
       Note that we cannot have any thread receiving a message while 
       checking the queues.   In case we do enqueue the message, we set
       the fields that will need to be valid BEFORE calling this routine
       (this is extra overhead ONLY in the case that the message was
       unexpected, which is already the higher-overhead case).
       */
    
    MPID_HANDLE_MUTEX_LOCK( rhandle );   
    rhandle->len	 = maxlen;
    rhandle->buf	 = buf;    
    MPID_HANDLE_COMPLETE_RESET( rhandle );
    rhandle->wait        = NULL;
    rhandle->test        = NULL;
    rhandle->finish      = NULL;
    MPID_HANDLE_MUTEX_UNLOCK( rhandle );
    
    MPID_Search_unexpected_queue_and_post( src_lrank, tag, context_id,  
					   rhandle, &dmpi_unexpected );
    if (dmpi_unexpected) {
      DEBUG_PRINT_MSG("R Found in unexpected queue");
      /* I need some syncro here, to avoid a call for NULL */
#ifdef GLOBAL_DEBUG
      marcel_printf("-----> IrecvContig : Waiting for Push ...\n"); 
#endif
      MPID_HANDLE_WAIT4PUSH( dmpi_unexpected );
#ifdef GLOBAL_DEBUG
      marcel_printf("<----  IrecvContig : Done.\n"); 
#endif
      
      DEBUG_TEST_FCN(dmpi_unexpected->push,"req->push");
      *error_code = (*(dmpi_unexpected->push))( rhandle, dmpi_unexpected );
      
#ifdef GLOBAL_DEBUG
      marcel_printf("IrecvContig : Push Done.\n");
#endif

      DEBUG_PRINT_MSG("R Exiting IrecvContig");
      /* This may or may not complete the message */
      return;
    }
    
    DEBUG_PRINT_MSG("R Exiting IrecvContig");
}


/*--------------------------------------------------------------------------------*/
int MPID_RecvIcomplete( 
		       MPI_Request request,
		       MPI_Status  *status,
		       int         *error_code)
{
  MPIR_RHANDLE *rhandle = &(request->rhandle);

  if ( MPID_HANDLE_TEST_4COMPLETE( rhandle ) )
    {
      if (status)
	{
	  *status   = rhandle->s;
	}
      *error_code   = rhandle->s.MPI_ERROR;
      return 1;
    }
  else
    {
      return 0;
    }
  }
/*--------------------------------------------------------------------*/
void MPID_RecvComplete( 
		       MPI_Request request,
		       MPI_Status  *status,
		       int         *error_code)
{
  MPIR_RHANDLE *rhandle = &(request->rhandle);
  
  MPID_HANDLE_MUTEX_LOCK( rhandle);
  MPID_HANDLE_MUTEX_LOCK_4COMPLETE( rhandle );
  if (!(MPID_HANDLE_TEST_4COMPLETE( rhandle )))
    {
      MPID_HANDLE_MUTEX_UNLOCK_4COMPLETE( rhandle );
      DEBUG_PRINT_MSG( "Entering while !rhandle->is_complete" );
      if( rhandle->wait)	
	{
	  MPID_HANDLE_MUTEX_UNLOCK( rhandle);
	  *error_code = (*(rhandle->wait))( rhandle );
	  MPID_HANDLE_MUTEX_LOCK( rhandle);
	}
      else 
	{
	  MPID_HANDLE_MUTEX_UNLOCK( rhandle);
	  *error_code = MPID_WaitForCompleteRecv( rhandle );
	  MPID_HANDLE_MUTEX_LOCK( rhandle);
	}
      DEBUG_PRINT_MSG( "Leaving while !rhandle->is_complete" );
    }
  else
    {
      MPID_HANDLE_MUTEX_UNLOCK_4COMPLETE( rhandle );
    } 
  
  if (status)
    { 
      *status = rhandle->s;
    }
  *error_code = rhandle->s.MPI_ERROR;
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );   
}
/*------------------------------------------------------------------*/
/* Temp fix for MPI_Status_set_elements, needed in Romio */
void MPID_Status_set_bytes( MPI_Status *status, int bytes )
{
    status->count = bytes;
}

#ifdef LAPI
/*\ JN: added to suck incoming messages when waiting for send to complete
\*/
int MPID_Check_Receives_While_Waiting()
{
int     lerr=0;
MPID_Device *dev = MPID_devset->dev_list;

      while (dev) {
           lerr = (*dev->check_device)( dev, MPID_NOTBLOCKING );
           if (lerr > 0) break;
           dev = dev->next;
      }
      return(lerr);
}
#endif







