/*
 *   $Id: testall.c,v 1.1.1.1 2003/07/30 15:35:53 lprylli Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      See COPYRIGHT in top-level directory.
 */

#include "mpiimpl.h"

#ifdef HAVE_WEAK_SYMBOLS

#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Testall = PMPI_Testall
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Testall  MPI_Testall
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Testall as PMPI_Testall
/* end of weak pragmas */
#endif

/* Include mapping from MPI->PMPI */
#define MPI_BUILD_PROFILING
#include "mpiprof.h"
/* Insert the prototypes for the PMPI routines */
#undef __MPI_BINDINGS
#include "binding.h"
#endif
#include "reqalloc.h"

/*@
    MPI_Testall - Tests for the completion of all previously initiated
    communications

Input Parameters:
+ count - lists length (integer) 
- array_of_requests - array of requests (array of handles) 

Output Parameters:
+ flag - (logical) 
- array_of_statuses - array of status objects (array of Status) 

Notes:
  'flag' is true only if all requests have completed.  Otherwise, flag is
  false and neither the 'array_of_requests' nor the 'array_of_statuses' is
  modified.

.N waitstatus

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_IN_STATUS

@*/
int MPI_Testall( 
		int count, 
		MPI_Request array_of_requests[], 
		int *flag, 
		MPI_Status array_of_statuses[] )
{
  int i, mpi_errno = MPI_SUCCESS;
  MPI_Request request;
  int nready;
  static char myname[] = "MPI_TESTALL";
  
  TR_PUSH(myname);
  
  /* It is a good thing that the receive requests contain the status object!
     We need this to save the status information in the case where not
     all of the requests have completed.
     
     Note that this routine forces some changes on the ADI test routines.
     It must be possible to test a completed request multiple times;
     once the "is_complete" field is set, the data must be saved until
     the request is explicitly freed.  That is, unlike the MPI tests,
     the ADI tests must be nondestructive.
   */
  nready = 0;

  for (i = 0; i < count; i++ ) 
    {
      request = array_of_requests[i];
      if (!request) 
	{
	  nready ++;
	  continue;
	}

      switch (request->handle_type) 
	{
	case MPIR_SEND:
	  MPID_HANDLE_MUTEX_LOCK(  &(request->shandle) );
	  if (MPID_SendRequestCancelled(request)) 
	    {
	      if (array_of_statuses) 
		{
		  array_of_statuses[i].MPI_TAG   = MPIR_MSG_CANCELLED; 
		  array_of_statuses[i].MPI_ERROR = MPI_SUCCESS;
		}
	      nready++; 
	    }
	  else 
	    {
	      MPID_HANDLE_MUTEX_LOCK_4COMPLETE(  &(request->shandle) );
	      if( MPID_HANDLE_TEST_4COMPLETE(  &(request->shandle) ) ||
		  MPID_SendIcomplete( request, &mpi_errno ))
		{
		  nready++;
		}
	      MPID_HANDLE_MUTEX_UNLOCK_4COMPLETE(  &(request->shandle) );    
	    }
	  MPID_HANDLE_MUTEX_UNLOCK(  &(request->shandle) );
	  break;
	  
	case MPIR_RECV:
	  MPID_HANDLE_MUTEX_LOCK( &(request->rhandle) );
	  if (request->rhandle.s.MPI_TAG == MPIR_MSG_CANCELLED) 
	    {
	      if (array_of_statuses) 
		{
		  array_of_statuses[i].MPI_TAG = MPIR_MSG_CANCELLED; 
		}
	      nready++; 
	    }
	  else 
	    {
	      MPID_HANDLE_MUTEX_LOCK_4COMPLETE(  &(request->rhandle) );
	      if( MPID_HANDLE_TEST_4COMPLETE( &(request->rhandle) ) ||
		  MPID_RecvIcomplete( request, (MPI_Status *)0, &mpi_errno ))
		{
		  nready++;
		}
	      MPID_HANDLE_MUTEX_UNLOCK_4COMPLETE(  &(request->rhandle) );
	    }
	  MPID_HANDLE_MUTEX_UNLOCK( &(request->rhandle) );
	  break;
	  
	case MPIR_PERSISTENT_SEND:
	  MPID_HANDLE_MUTEX_LOCK(  &(request->persistent_shandle.shandle) );
	  MPID_HANDLE_MUTEX_LOCK_4COMPLETE( &(request->persistent_shandle.shandle) );
	  if ((request->persistent_shandle.active) && 
	      !(MPID_HANDLE_TEST_4COMPLETE( &(request->persistent_shandle.shandle) )))
	    {
	      if (MPID_SendIcomplete( request, &mpi_errno ))
		{
		  nready++;
		}
	    }
	  else
	    { 
	      nready++;
	    }
	  MPID_HANDLE_MUTEX_UNLOCK_4COMPLETE( &(request->persistent_shandle.shandle) );
	  MPID_HANDLE_MUTEX_UNLOCK(  &(request->persistent_shandle.shandle) );
	  break;
	  
	case MPIR_PERSISTENT_RECV:
	  MPID_HANDLE_MUTEX_LOCK( &(request->persistent_rhandle.rhandle) );
	  MPID_HANDLE_MUTEX_LOCK_4COMPLETE( &(request->persistent_rhandle.rhandle) );
	  if ((request->persistent_rhandle.active) &&
	      !(MPID_HANDLE_TEST_4COMPLETE( &(request->persistent_rhandle.rhandle) )))
	    {
	      if (MPID_RecvIcomplete( request, (MPI_Status *)0, &mpi_errno ))
		{
		  nready++;
		}
	    }
	  else
	    { 
	      nready++;
	    }
	  MPID_HANDLE_MUTEX_UNLOCK_4COMPLETE( &(request->persistent_rhandle.rhandle) );	
	  MPID_HANDLE_MUTEX_UNLOCK( &(request->persistent_rhandle.rhandle) );
	  break;
	}


      if (mpi_errno) 
	{
	  MPIR_Set_Status_error_array( array_of_requests, count, i, 
				       mpi_errno, array_of_statuses );
	  mpi_errno = MPI_ERR_IN_STATUS;
	  TR_POP;
	  MPIR_RETURN(MPIR_COMM_WORLD, mpi_errno, myname );
	}
    }
  
  *flag = (nready == count);
  
  /* Because a request may have completed with an error (such as 
     MPI_ERR_TRUNCATE), we need to check here as well */
  if (nready == count) 
    {
      for (i=0; i<count; i++) 
	{
	  request = array_of_requests[i];
	    if (!request) {
	      /* See MPI Standard, 3.7 */
	     if (array_of_statuses) 
	       {
		 array_of_statuses[i].MPI_TAG	 = MPI_ANY_TAG;
		 array_of_statuses[i].MPI_SOURCE = MPI_ANY_SOURCE;
		 array_of_statuses[i].MPI_ERROR	 = MPI_SUCCESS;
		 MPID_ZERO_STATUS_COUNT(&array_of_statuses[i]);
	       }
	     continue;
	    }
	    
	    switch (request->handle_type) 
	      {
	      case MPIR_SEND:
		if ((array_of_statuses && 
		     array_of_statuses[i].MPI_TAG != MPIR_MSG_CANCELLED) ||
		    !MPID_SendRequestCancelled(request))
		  
		  {
		    MPIR_FORGET_SEND( &request->shandle );
		    MPID_SendFree( array_of_requests[i] );
		    array_of_requests[i] = 0;
		  }
		break;
		
	      case MPIR_RECV:
		MPID_HANDLE_MUTEX_LOCK( &(request->rhandle) );
		if (request->rhandle.s.MPI_TAG != MPIR_MSG_CANCELLED) 
		  {
		    if (request->rhandle.s.MPI_ERROR) 
		      {
			mpi_errno = request->rhandle.s.MPI_ERROR;
		      }
		    if (array_of_statuses) 
		      {
			array_of_statuses[i] = request->rhandle.s;
		      }
		    MPID_HANDLE_MUTEX_UNLOCK( &(request->rhandle) );
		    MPID_RecvFree( array_of_requests[i] );
		    array_of_requests[i] = 0;
		  }
		else
		  {
		    MPID_HANDLE_MUTEX_UNLOCK( &(request->rhandle) );
		  }
		break;
		
	      case MPIR_PERSISTENT_SEND:
		MPID_HANDLE_MUTEX_LOCK( &(request->persistent_shandle.shandle) );
		if (request->persistent_shandle.active) 
		  {
		    if (array_of_statuses){
		      array_of_statuses[i].MPI_ERROR = 
			MPID_SendRequestErrval(&request->persistent_shandle.shandle);
		    }
		    request->persistent_shandle.active = 0;
		  }
		else {
		  /* See MPI Standard, 3.7 */
		  /* Thanks to mechi@terra.co.il for this fix */
		  if (array_of_statuses)
		    {
		      if (MPID_SendRequestCancelled( &request->persistent_shandle) )
			{
			  array_of_statuses[i].MPI_TAG  = MPIR_MSG_CANCELLED;
			}
		      else
			{
			  array_of_statuses[i].MPI_TAG  = MPI_ANY_TAG;
			}
		      array_of_statuses[i].MPI_SOURCE   = MPI_ANY_SOURCE;
		      array_of_statuses[i].MPI_ERROR    = MPI_SUCCESS;
		      MPID_ZERO_STATUS_COUNT(&array_of_statuses[i]);
		    }
		}
		MPID_HANDLE_MUTEX_UNLOCK( &(request->persistent_shandle.shandle) );
		break;
		
	      case MPIR_PERSISTENT_RECV:
		MPID_HANDLE_MUTEX_LOCK( &(request->persistent_rhandle.rhandle) );
		if (request->persistent_rhandle.active) 
		  {
		    if (array_of_statuses) 
		      {
			array_of_statuses[i] = 
			  request->persistent_rhandle.rhandle.s;
		      }
		    mpi_errno = request->persistent_rhandle.rhandle.s.MPI_ERROR;
		    request->persistent_rhandle.active = 0;
		  }
		else {
		  /* See MPI Standard, 3.7 */
		  /* Thanks to mechi@terra.co.il for this fix */
		  if (array_of_statuses) 
		    {
		      if (request->persistent_rhandle.rhandle.s.MPI_TAG ==
			  MPIR_MSG_CANCELLED) 
			{
			  array_of_statuses[i].MPI_TAG = MPIR_MSG_CANCELLED;
			}
		      else
			{
			  array_of_statuses[i].MPI_TAG = MPI_ANY_TAG;
			}
		      array_of_statuses[i].MPI_SOURCE  = MPI_ANY_SOURCE;
		      array_of_statuses[i].MPI_ERROR   = MPI_SUCCESS;
		      MPID_ZERO_STATUS_COUNT(&array_of_statuses[i]);
		    }
		}
		MPID_HANDLE_MUTEX_UNLOCK( &(request->persistent_rhandle.rhandle) );
		break;
	      }
	    if (mpi_errno) {
	      if (array_of_statuses) 
		{
		  MPIR_Set_Status_error_array( array_of_requests, count, i, 
					       mpi_errno, array_of_statuses );
		}
	      mpi_errno = MPI_ERR_IN_STATUS;
	      TR_POP;
	      MPIR_RETURN(MPIR_COMM_WORLD, mpi_errno, myname );
	    }
	}
    }
  TR_POP;
  MPIR_RETURN(MPIR_COMM_WORLD, mpi_errno, myname );
}













